import logging
from abc import ABC
from typing import List, Optional

from src.pydisney.Auth import Auth
from src.pydisney.Config import APIConfig
from src.pydisney.models.Season import Season
from src.pydisney.utils.parser import safe_get

logger = logging.getLogger("pydisney")
logger = logger.getChild("Hit")


class Hit(ABC):

    def __init__(self, data):
        self.id: Optional[str] = None
        self.title: Optional[str] = None
        self.brief_desc: Optional[str] = None
        self.medium_desc: Optional[str] = None
        self.full_desc: Optional[str] = None
        self.genres: Optional[List[str]] = None
        self.rating: Optional[str] = None
        self.rating: Optional[str] = None
        self.artwork = None
        self.startYear: Optional[str] = None

        self.__seasons: List[Season] = []
        self.__is_movie: Optional[bool] = None
        self.__endYear: Optional[str] = None
        self.__credits: Optional[dict] = None
        self.__release_date: Optional[str] = None
        self.__length: Optional[int] = None
        self._parse_data(data)

    def _parse_data(self, data):

        self.id = data["id"]
        self.title = data["visuals"]["title"]
        self.brief_desc = safe_get(data, ["visuals", "description", "brief"])
        self.medium_desc = safe_get(data, ["visuals", "description", "medium"])
        self.full_desc = safe_get(data, ["visuals", "description", "full"])
        self.genres = safe_get(data, ["visuals", "metastringParts", "genres", "values"], [])
        self.rating = safe_get(data, ["visuals", "metastringParts", "ratingInfo", "rating", "text"])
        self.startYear = safe_get(data, ["visuals", "metastringParts", "releaseYearRange", "startYear"])
        self.artwork = safe_get(data, ["visuals", "artwork"])

    @staticmethod
    def parse_hits(data):
        hits = []
        for element in data:
            hits.append(Hit(element))
        return hits

    @property
    def is_movie(self):
        if self.__is_movie is None:
            self._get_more_data()
        return self.__is_movie

    @property
    def length(self):
        if not self.is_movie:
            raise ValueError("Series has no length property")
        if self.__length is None:
            self._get_more_data()
        return self.__length

    @property
    def seasons(self):
        if self.is_movie:
            raise ValueError("Movie has no seasons property")
        if not self.__seasons:
            self._get_more_data()
        return self.__seasons

    @property
    def credits(self):
        if not self.__credits:
            self._get_more_data()
        return self.__credits

    @property
    def endYear(self):
        if not self.__endYear:
            self._get_more_data()
        return self.__endYear

    @property
    def release_date(self):
        if not self.__release_date:
            self._get_from_old_api()
        return self.__release_date

    def _get_more_data(self) -> None:
        url = f"https://disney.api.edge.bamgrid.com/explore/v1.7/page/entity-{self.id}"
        res = Auth.make_request("GET", url)
        self.__is_movie = True
        for container in res["data"]["page"]["containers"]:
            if container["type"] == "episodes":
                self.__is_movie = False
                for index, season_data in enumerate(container["seasons"]):
                    self.__seasons.append(Season(season_data, index))

            elif container["type"] == "set":
                # ignoring suggested and extras for now
                pass

            elif container["type"] == "details":
                data = safe_get(container, ["visuals", "credits"])
                if data:
                    formatted_data = {}
                    for entry in data:
                        heading = entry['heading']
                        items = [item['displayText'] for item in entry['items']]
                        formatted_data[heading] = items
                    self.__credits = formatted_data
                self.__endYear = safe_get(container, ["visuals", "metastringParts", "releaseYearRange", "endYear"], ignoreError=True)
                self.__length = safe_get(container, ["visuals", "duration", "runtimeMs"])

    def _get_from_old_api(self) -> None:
        logger.info("Attempting to fetch data from old api")

        res = Auth.make_request("GET",
                                f"https://disney.content.edge.bamgrid.com/svc/search/disney/version/5.1/region/{APIConfig.region}/audience/k-false,l-true/maturity/1850/language/{APIConfig.language}/queryType/ge/pageSize/1/query/{self.title}")
        item = res["data"]["search"]["hits"][0]["hit"]

        self.__release_date = item["releases"][0]["releaseDate"]
        self.__encodedSeriesId = item["encodedSeriesId"]

    def _monkey_patch(self) -> None:
        pass
        # todo
        # res = Auth.make_request("GET",
        #     f"https://disney.content.edge.bamgrid.com/svc/content/DmcSeriesBundle/version/5.1/region/{APIConfig.region}/audience/false/maturity/1850/language/{APIConfig.language}/encodedSeriesId/106A3s2Armta")
        # print(res)
        # res = Auth.make_get_request(
        #     f"https://disney.content.edge.bamgrid.com/svc/content/DmcSeriesBundle/version/5.1/region/{APIConfig.region}/audience/false/maturity/1850/language/{APIConfig.language}/encodedSeriesId/{self.encoded_series_id}")
        #
        # res_json = res.json()["data"]["DmcSeriesBundle"]
        # self._full_description = res_json["series"]["text"]["description"]["full"]["series"]["default"]["content"]
        # self._medium_description = res_json["series"]["text"]["description"]["medium"]["series"]["default"]["content"]
        # self._brief_description = res_json["series"]["text"]["description"]["brief"]["series"]["default"]["content"]
        #
        # actors, directors, producers, creators = parse_participants(res_json["series"]["participant"])
        # self._cast = actors
        # self._directors = directors
        # self._producers = producers
        # self._creators = creators
        #
        # seasons = []
        # for season_json in res_json["seasons"]["seasons"]:
        #     season_id = season_json["seasonId"]
        #     number = season_json["seasonSequenceNumber"]
        #
        #     season = Season(season_id=season_id, number=number)
        #
        #     season.release_date = season_json["releases"][0]["releaseDate"]
        #     season.release_year = season_json["releases"][0]["releaseYear"]
        #     season.rating = season_json["ratings"][0]["value"]
        #     season.encoded_series_id = season_json["encodedSeriesId"]
        #     season.series_id = season_json["seriesId"]
        #
        #     seasons.append(season)
        #
        # return seasons

    def __str__(self):
        return f"[Title={self.title}]"

    def __repr__(self):
        return self.title



import logging

from ..models.Profile import Profile

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

def parse_profile(profile_json):
    profile_id = profile_json["id"]
    name = profile_json["name"]
    kids_mode = profile_json["attributes"]["kidsModeEnabled"]
    is_default = profile_json["attributes"]["isDefault"]
    av_id = profile_json["attributes"]["avatar"]["id"]
    av_user_selected = profile_json["attributes"]["avatar"]["userSelected"]
    prf = Profile(profile_id=profile_id, name=name, kids_mode=kids_mode, is_default=is_default, av_id=av_id,
                  av_user_selected=av_user_selected)

    prf.language_preferences.app = profile_json["attributes"]["languagePreferences"]["appLanguage"]

    prf.language_preferences.playback = profile_json["attributes"]["languagePreferences"]["playbackLanguage"]

    prf.language_preferences.subtitle = profile_json["attributes"]["languagePreferences"]["subtitleLanguage"]
    prf.language_preferences.subs_enabled = profile_json["attributes"]["languagePreferences"]["subtitlesEnabled"]
    return prf


def find_pagination_values(data):
    """
    Recursively walks through a JSON object to find the first pagination structure and return its values.

    :param data: The JSON data (dict or list).
    :return: A JSON object containing the first pagination structure found.
    """
    pagination_keys = {"hasMore", "hasPrev", "currentOffset"}

    if isinstance(data, dict):
        # Check if current dictionary contains pagination keys
        if pagination_keys.issubset(data.keys()):
            return {
                "hasMore": data["hasMore"],
                "hasPrev": data["hasPrev"],
                "currentOffset": data["currentOffset"],
                "totalCount": data.get("totalCount")
            }

        # Recurse into each value in the dictionary
        for value in data.values():
            result = find_pagination_values(value)
            if result:
                return result

    elif isinstance(data, list):
        # Recurse into each item in the list
        for item in data:
            result = find_pagination_values(item)
            if result:
                return result

    return None


def safe_get(d, keys, default=None, ignoreError=False):
    """Safely retrieve nested dictionary keys."""
    current = d
    try:
        for key in keys:
            current = current[key]
        return current
    except (KeyError, TypeError):
        if not ignoreError:
            logger.warning(f"Missing key: {''.join(f'[{key}]' for key in keys)}")
        return default





class LanguagePreferences:
    def __init__(self):
        self.playback = None
        self.subtitle = None
        self.subs_enabled = None
        self.app = None

    def __str__(self):
        return self.app



from enum import Enum


class Language(Enum):
    Czech = "cs"
    Danish = "da"
    German = "de"
    English_UK = "en-GB"
    English_US = "en"
    Spanish_ES = "es-ES"
    Spanish_LatinAmerica = "es-419"
    French_FR = "fr-FR"
    French_CA = "fr-CA"
    Italian = "it"
    Hungarian = "hu"
    Dutch = "nl"
    Norwegian = "no"
    Polish = "pl"
    Portuguese_PT = "pt-PT"
    Portuguese_BR = "pt-BR"
    Romanian = "ro"
    Slovak = "sk"
    Finnish = "fi"
    Swedish = "sv"
    Turkish = "tr"
    Greek = "el"
    Korean = "ko"
    Chinese_Simplified = "zh-Hans"
    Chinese_Traditional = "zh-Hant"
    Chinese_HK = "zh-HK"
    Japanese = "ja"




class Account:
    def __init__(self, account_id, email, is_email_verified, country, created_at):
        self.id = account_id
        self.email = email
        self.is_email_verified = is_email_verified
        self.country = country
        self.created_at = created_at

    def __str__(self):
        return f"Account[Email={self.email}, Id={self.id}, Country={self.country}, Created_at={self.created_at}"

    def __repr__(self):
        return self.id



from requests import Response


class AuthException(Exception):
    def __init__(self, message: Response):
        self.message = message
        super().__init__(f"{self.message.text}\n{self.message.status_code}")

class ApiException(Exception):
    def __init__(self, response: Response):
        self.message = response
        super().__init__(f"{self.message.text}\n{self.message.status_code}")

class GraphqlException(Exception):
    def __init__(self, response: list):
        self.message = response
        super().__init__(str(response))

class FFmpegException(Exception):
    def __init__(self):
        super().__init__("Unknown ffmpeg error occured")



from typing import Optional

from src.pydisney.utils.parser import safe_get


class Episode:
    def __init__(self, data):
        self.artwork: Optional[dict] = None
        self.durationMs: Optional[int] = None
        self.brief_description: Optional[str] = None
        self.medium_description: Optional[str] = None
        self.full_description: Optional[str] = None
        self.episodeNumber: Optional[int] = None
        self.episodeTitle: Optional[str] = None
        self.fullEpisodeTitle: Optional[str] = None
        self.rating: Optional[dict] = None

        self._parse_data(data)

    def _parse_data(self, data) -> None:
        self.artwork = safe_get(data, ["visuals", "artwork"])
        self.durationMs = safe_get(data, ["visuals", "durationMs"])
        self.brief_description = safe_get(data, ["visuals", "description", "brief"])
        self.medium_description = safe_get(data, ["visuals", "description", "medium"])
        self.full_description = safe_get(data, ["visuals", "description", "full"])

        self.episodeNumber = safe_get(data, ["visuals", "episodeNumber"])
        self.episodeTitle = safe_get(data, ["visuals", "episodeTitle"])
        self.fullEpisodeTitle = safe_get(data, ["visuals", "fullEpisodeTitle"])

        self.rating = safe_get(data, ["visuals", "metastringParts", "ratingInfo", "rating", "text"])

    def __str__(self):
        return f"[Title={self.fullEpisodeTitle}, Episode={self.episodeNumber}]"

    def __repr__(self):
        return self.fullEpisodeTitle



import json
import re
from datetime import timedelta, datetime

from ..Config import APIConfig


def rename_filename(filename):
    """Fix invalid character from title"""

    filename = (
        filename.replace(" ", ".")
        .replace("'", "")
        .replace('"', "")
        .replace(",", "")
        .replace("-", "")
        .replace(":", ".")
        .replace("’", "")
        .replace('"', '')
        .replace("-.", ".")
        .replace(".-.", ".")
    )
    filename = re.sub(" +", ".", filename)
    for _ in range(5):
        filename = re.sub(r"(\.\.)", ".", filename)
    filename = re.sub(r'[/\\:|<>"?*\0-\x1f]|^(AUX|COM[1-9]|CON|LPT[1-9]|NUL|PRN)(?![^.])|^\s|[\s.]$',
                      "", filename[:255], flags=re.IGNORECASE)

    return filename


def update_file():
    with open("token.json", "w") as file:
        current_time = datetime.now()
        expiration_time = current_time + timedelta(hours=4)
        token_data = {
            "token": APIConfig.token,
            "refresh": APIConfig.refresh,
            "expiration_time": expiration_time.strftime("%Y-%m-%d %H:%M:%S")
        }
        json.dump(token_data, file)



from typing import Optional, List

from .Episode import Episode
from ..Auth import Auth


class Season:
    def __init__(self, data, number):
        self.id: Optional[str] = None
        self.totalEpisodesCount: Optional[int] = None
        self.name: Optional[str] = None
        self.number: Optional[int] = number

        self.__episodes: Optional[List[Episode]] = []
        self._parse_data(data)

    def _parse_data(self, data):
        self.id = data["id"]
        self.totalEpisodesCount = data["visuals"]["episodeCountDisplayText"]
        self.name = data["visuals"]["name"]

    @property
    def episodes(self):
        if not self.__episodes:
            self._get_more_data()
        return self.__episodes

    def _get_more_data(self):
        url = f"https://disney.api.edge.bamgrid.com/explore/v1.7/season/{self.id}"
        res = Auth.make_pagination_request("GET", url)
        for page in res:
            for element in page["data"]["season"]["items"]:
                self.__episodes.append(Episode(element))
        return self.__episodes

    def __str__(self):
        return f"[name={self.name}]"

    def __repr__(self):
        return self.name



from .Language import Language
from .Rating import Rating
from ..Auth import Auth
from ..Config import APIConfig
from ..models.LanguagePreferences import LanguagePreferences


class Profile:
    def __init__(self, profile_id, name, kids_mode, is_default, av_id, av_user_selected):
        self.id = profile_id
        self.name = name
        self.kids_mode = kids_mode
        self.avatar = {"id": av_id, "user_selected": av_user_selected}
        self.is_default = is_default
        self.language_preferences = LanguagePreferences()

    def _get_action_grant(self) -> str:
        graphql_mutation = {
            "query": """
                    mutation authenticate($input: AuthenticateInput!) {
                      authenticate(authenticate: $input) {
                        actionGrant
                      }
                    }
                  """,
            "variables": {
                "input": {
                    "email": APIConfig.auth._email,
                    "password": APIConfig.auth._password,
                    "reasons": [
                        "UpdateProfileMaturityRating"
                    ]
                }
            },
            "operationName": "authenticate"
        }
        res = Auth.make_request("POST", "https://disney.api.edge.bamgrid.com/v1/public/graphql", data=graphql_mutation)
        return res["data"]["authenticate"]["actionGrant"]

    def set_profile_maturity_rating(self, rating: Rating) -> None:
        actionGrant = self._get_action_grant()
        graphql_mutation = {
            "query": """
            mutation updateProfileMaturityRatingWithActionGrant($input: UpdateProfileMaturityRatingWithActionGrantInput!) {
              updateProfileMaturityRatingWithActionGrant(updateProfileMaturityRatingWithActionGrant: $input) {
                accepted
              }
            }
          """,
            "variables": {
                "input": {
                    "actionGrant": actionGrant,
                    "contentMaturityRating": rating.value,
                    "profileId": self.id,
                    "ratingSystem": "DisneyPlusEMEA"
                }
            },
            "operationName": "updateProfileMaturityRatingWithActionGrant"
        }
        Auth.make_request("POST", "https://disney.api.edge.bamgrid.com/v1/public/graphql", data=graphql_mutation)

    def set_profile_language(self, language: Language):
        graphql_mutation = {
            "query": """
            mutation updateProfileAppLanguage($input: UpdateProfileAppLanguageInput!) {
                updateProfileAppLanguage(updateProfileAppLanguage: $input) {
                    accepted
                }
            }
            """,
            "variables": {
                "input": {
                    "appLanguage": language.value,
                    "profileId": "3f596128-1035-4872-adea-1eeddf973625"
                }
            },
            "operationName": "updateProfileAppLanguage"
        }
        Auth.make_request("POST", "https://disney.api.edge.bamgrid.com/v1/public/graphql", data=graphql_mutation)

    def __str__(self):
        return f"Profile({self.name})"

    def __repr__(self):
        return self.name


